package dod;

import dod.game.CommandException;
import dod.game.CompassDirection;
import dod.game.GameLogic;
import dod.game.Player;
import dod.game.PlayerListener;

public abstract class PlayerGuide implements PlayerListener {

	private GameLogic game;
	private ClientServerController ChildCSC;

	// Handled by ClientThread as this uses network
	public abstract void run();
	
	// Handled by ClientThread as this uses network
	abstract protected void outputMessage(String message);
	
    /**
     * Sends a message to the player from the game.
     * 
     * @ param message The message to be sent
     */
    @Override
    public void sendMessage(String message) {
	outputMessage("MESSAGE " + message);
    }
    
    /**
     * Sends a message to the player from the game.
     * 
     * @ param message The message to be sent
     */
    @Override
    public void sendChatMessage(String message) {
	outputMessage("CHAT" + System.getProperty("line.separator") + message);
    }
    
    /**
     * Informs the user of the beginning of a player's turn
     */
    @Override
    public void startTurn() {
	outputMessage("STARTTURN");
    }
    
    @Override
    public void change() {
	outputMessage("CHANGE");
    }

    /**
     * Informs the user of the end of a player's turn
     */
    @Override
    public void endTurn() {
	outputMessage("ENDTURN");
    }

    /**
     * Informs the user that the player has won
     */
    @Override
    public void win() {
	outputMessage("WIN");
    }
    
    @Override 
    public void lose()
    {
    	outputMessage("LOSE");
    }

    /**
     * Informs the user that the player's hit points have changed
     */
    @Override
    public void hpChange(int value) {
	outputMessage("HITMOD " + value);
    }

    /**
     * Informs the user that the player's gold count has changed
     */
    @Override
    public void treasureChange(int value) {
	outputMessage("TREASUREMOD " + value);
    }
    
    protected void checkTurn(ClientThread thread){
		ChildCSC.checkNewTurn(game, thread);
	}
    
	public void addGame(ClientThread thread) {
		ChildCSC.addGame(this,game,thread);
		outputMessage("GOAL " + getGoal());
	}
	
	public int getGoal(){
		return ChildCSC.getGoal(game);
	}
	
    /**
     * Processes a text command from the user.
     * 
     * @param commandString
     *            the string containing the command and any argument
     */
    public void processCommand(String commandString, ClientThread thread) {
		commandString = commandString.toUpperCase();
		// Process the command string e.g. MOVE N
		final String commandStringSplit[] = commandString.split(" ", 2);
		final String command = commandStringSplit[0];
		final String arg = ((commandStringSplit.length == 2) ? commandStringSplit[1]: null);
		try {
		    processCommandAndArgument(command, arg, thread);
		} catch (CommandException e) {
		    outputMessage("FAIL " + e.getMessage());
		}
    }
    
	public void setCSC(ClientServerController CSC)
	{
		ChildCSC = CSC;
	}
	
	public void setGame(GameLogic game) {
		this.game = game;
	}
	
	/**
	 * Tells the game logic a player has left and
	 * will remove them from the map and turns list
	 */
	public void UserConnectionGone(ClientThread thread)
	{
		// kill the player in the game
		ChildCSC.die(game, thread);
		// remove them from the turns list
		ChildCSC.playerleft(game, thread);
		
	}

    /**
     * Processes the command and an optional argument
     * 
     * @param command
     *            the text command
     * @param arg
     *            the text argument (null if no argument)
     * @throws CommandException
     */
    private void processCommandAndArgument(String command, String arg, ClientThread thread) throws CommandException {
		if (command.equals("HELLO")) { // HELLO
		    if (arg == null) {throw new CommandException("HELLO needs an argument");}
		    String name = sanitiseMessage(arg);
		    ChildCSC.clientHello(game, name, thread);
		    outputMessage("HELLO " + name);
		} else if (command.equals("LOOK")) { // LOOK
		    if (arg != null) {throw new CommandException("LOOK does not take an argument");}
		    outputMessage("LOOKREPLY" + System.getProperty("line.separator") + ChildCSC.clientLook(game, thread));
		} else if (command.equals("PICKUP")) { // PICKUP
		    if (arg != null) {throw new CommandException("PICKUP does not take an argument");}
		    ChildCSC.clientPickup(game, thread);
		    outputSuccess();
		} else if (command.equals("MOVE")) { // MOVE
		    if (arg == null) {throw new CommandException("MOVE needs a direction");}
		    ChildCSC.clientMove(getDirection(arg), game, thread);
		    outputSuccess();
			//check to see if this has changed someones map
		    ChildCSC.checkChange(game, thread);
			// Notify the client of the success
		    ChildCSC.advanceTurn(game, thread);
		} else if (command.equals("CHAT")) { // CHAT
		    if (arg == null) {throw new CommandException("CHAT needs a Message");}
		    ChildCSC.clientCHAT(game, arg);
		} else if (command.equals("ATTACK")) { // ATTACK
		    // We need to know which direction to move in.
		    if (arg == null) {throw new CommandException("ATTACK needs a direction");}
		    ChildCSC.clientAttack(getDirection(arg), game, thread);
		    outputSuccess();
		} else if (command.equals("ENDTURN")) { // ENDTURN
			ChildCSC.clientEndTurn(game, thread);
		} else if (command.equals("SHOUT")) { // SHOUT
		    // Ensure they have given us something to shout.
		    if (arg == null) {throw new CommandException("need something to shout");}
		    ChildCSC.clientShout(sanitiseMessage(arg), game, thread);
		    outputSuccess();
		} else {
		    // If it is none of the above then it must be a bad command.
		    throw new CommandException("Invalid Command");
		}
    }

    private CompassDirection getDirection(String string)throws CommandException {
    	try {
    	    return CompassDirection.fromString(string);
    	} catch (final IllegalArgumentException e) {
    	    throw new CommandException("invalid direction");
    	}
    }
    
    /**
     * Sanitises the given message - there are some characters that we can put
     * in the messages that we don't want in other stuff that we sanitise.
     * 
     * @param s
     *            The message to be sanitised
     * @return The sanitised message
     */
    private static String sanitiseMessage(String s) {
	return sanitise(s, "[a-zA-Z0-9-_ \\.,:!\\(\\)#]");
    }

    /**
     * Strip out anything that isn't in the specified regex.
     * 
     * @param s
     *            The string to be sanitised
     * @param regex
     *            The regex to use for sanitisiation
     * @return The sanitised string
     */
    private static String sanitise(String s, String regex) {
	String rv = "";

	for (int i = 0; i < s.length(); i++) {
	    final String tmp = s.substring(i, i + 1);

	    if (tmp.matches(regex)) {
		rv += tmp;
	    }
	}

	return rv;
    }

    /**
     * Sends a success message in the event that a command has succeeded
     */
    private void outputSuccess() {
	outputMessage("SUCCESS");
    }

}
