package dod.game;

import java.io.FileNotFoundException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;

import dod.ClientThread;
import dod.game.items.GameItem;

public class GameLogic {
	
	private static Map map;

	private boolean playerWon = false;

	// List of players and a List of Players Names
	private ArrayList<Player> userList = new ArrayList<Player>();
	private ArrayList<String> userNames = new ArrayList<String>();
	private HashMap<ClientThread, Player> playerThreadMap = new HashMap<ClientThread, Player>();

	// used to keep a hold of the number of player keeping a unique name for
	// them
	private int playCount = 0;

	/**
	 * Constructor that specifies the map which the game should be played on.
	 * 
	 * @param mapFile
	 *            The name of the file to load the map from.
	 * @throws FileNotFoundException
	 *             , ParseException
	 */
	public GameLogic(String mapFile) throws FileNotFoundException, ParseException {
		map = new Map(mapFile);
		// Check if there is enough gold to win
		if (map.remainingGold() < map.getGoal()) {
			throw new IllegalStateException("There isn't enough gold on this map for you to win");
		}
	}
	
	/**
	 * Adds a new player to the game. For now, only one player can be added.
	 * 
	 * @param listener
	 *            the PlayerListener which will listen on behalf of that player,
	 *            so messages can be sent to the player
	 * @param thread
	 *            The current player used by the multithreading for clients
	 */
	public synchronized void addPlayer(PlayerListener listener, ClientThread thread) {
		// creates the uniques name
		String playername = "Player " + playCount;
		// sets the client thread's player varibale = to the player
		Player player = new Player(playername, generateRandomStartLocation(), listener, thread, map);
		// used for unique name
		playCount++;
		// place the player in the map to be seen
		player.placeThePlayer();
		// add the player to the thread list and names list
		userList.add(player);
		userNames.add(playername);
		setPlayerThreadMap(thread, player);
	}
	
	public Player getPlayerThreadMap(ClientThread thread) {
		return playerThreadMap.get(thread);
	}

	public void setPlayerThreadMap(ClientThread thread, Player player) {
		playerThreadMap.put(thread, player);
	}
	
	/**
	 * @param Message
	 *            - Html message to be sent in chat
	 * @param thread
	 *            - User thread who sent message Sends a chat message to
	 *            everyone in the game
	 */
	public void clientCHAT(String Message) {
		// for all user in game, call player chat message
		for (int user = 0; user < userList.size(); user++) {
			userList.get(user).sendChatMessage(Message);
		}

	}
	
	/**
	 * Zero's the Ap of the player creates a new turn for the next player
	 * removes the player form the thread list
	 * 
	 * @param thread
	 *            The current player used by the multithreading for clients
	 */
	public void playerleft(Player player) {
		player.zeroAP();
		checkNewTurn(player);
		userList.remove(player);
	}

	/**
	 * Will work out which player gets the next turn to use their AP points
	 * 
	 * @param thread
	 *            The current player used by the multithreading for clients
	 */
	public void checkNewTurn(Player player) {
		// for all the users
		for (int user = 0; user < userList.size(); user++) {
			// if the user is still going
			if (userList.get(user).remainingAp() > 0) {
				System.out.println("turn in progress");
				break;
			// if you have 0 ap and its your thread
			} else if ((userList.get(user).remainingAp() == 0)
					&& (userList.get(user) == player)) {
				// start the first user as it has hit the end of the list
				if (user + 1 == userList.size()) {
					System.out.println("Start first user");
					startTurn(userList.get(0));
					break;
					// Start the next user in the list after you
				} else {
					System.out.println("Start next user");
					startTurn(userList.get(user + 1));
					break;
				}
			}
		}
	}
	
	/**
	 * Will get the players gold and drop it on the tile underneath the player
	 * 
	 * @param thread
	 *            The current player used by the multithreading for clients
	 */
	public void die(Player player) {
		int goldValue = player.getGold();
		if (goldValue > 0) {
			if (player.getTempTile().toChar() == 'G') {
				map.setMapCellGold(player.getLocation(),
						player.getGold() + player.getTempTile().getValue());
			} else if (player.getTempTile().toChar() == 'E') {
				// find a random empty tile and place gold
				while (true) {
					Location temp = generateRandomStartLocation();
					if (map.getMapCell(temp).toChar() != 'E') {
						// if it has a G add the gold to this tile
						if (map.getMapCell(temp).toChar() == 'G') {
							map.setMapCellGold(temp, player.getGold() + map.getMapCell(temp).getValue());
						} else {
							// overwrite with gold amount on tile
							map.setMapCellGold(temp, player.getGold());
						}
						break;
					}
				}
			} else {
				// Just drop player gold on current tile
				map.setMapCellGold(player.getLocation(),player.getGold());
			}
		} else {
			// removes their 'P' from the map
			removePlayerFromMap(player);
		}
		// check to see if this has changed someones map
		checkChange(player);
	}
	
	/**
	 * Handles the client message HELLO
	 * 
	 * @param newName
	 *            the name of the player to say hello to
	 * @return the message to be passed back to the command line
	 * @throws CommandException
	 */
	public void clientHello(String newName, Player player) throws CommandException {
		assertPlayerExists(player);
		// Change the player name and then say hello to them
		player.setName(newName,this);
	}
	
	/**
	 * Handles the client message LOOK Shows the portion of the map that the
	 * player can currently see.
	 * 
	 * @return the part of the map that the player can currently see.
	 */
	public synchronized String clientLook(Player player) {
		assertPlayerExists(player);

		// Work out how far the player can see
		final int distance = player.lookDistance();

		String lookReply = "";
		// Iterate through the rows.
		for (int rowOffset = -distance; rowOffset <= distance; ++rowOffset) {
			String line = "";

			// Iterate through the columns.
			for (int colOffset = -distance; colOffset <= distance; ++colOffset) {

				// Work out the location
				final Location location = player.getLocation().atOffset(colOffset, rowOffset);

				char content = '?';
				if (!player.canSeeTile(rowOffset, colOffset)) {
					// It's outside the FoV so we don't know what it is.
					content = 'X';
				} else if (!map.insideMap(location)) {
					// It's outside the map, so just call it a wall.
					content = '#';
				} else {
					if (map.getMapCell(location) == player.getCurrentTile()) {
						System.out.println("Player was found");
						content = player.getTempTile().toChar();
					} else {
						content = map.getMapCell(location).toChar();
					}
				}

				// Add to the line
				line += content;
			}

			// Send a line of the look message
			lookReply += line + System.getProperty("line.separator");
		}

		return lookReply;
	}
	
	/**
	 * Handles the client message MOVE
	 * 
	 * Move the player in the specified direction - assuming there isn't a wall
	 * in the way
	 * 
	 * @param direction
	 *            The direction (NESW) to move the player
	 * @return An indicator of the success or failure of the movement.
	 * @throws CommandException
	 */
	public synchronized void clientMove(CompassDirection direction, Player player) throws CommandException {
		assertPlayerExists(player);
		ensureNoWinner();
		try {
			assertPlayerAP(player);
		} catch (IllegalStateException e) {
			player.sendMessage("You have 0 ap");
			throw new CommandException("Player has 0 ap");
		}

		// Work out where the move would take the player
		final Location location = player.getLocation().atCompassDirection(direction);

		// Ensure that the movement is within the bounds of the map and not
		// into a wall
		if (!map.insideMap(location) || !map.getMapCell(location).isWalkable()) {
			player.sendMessage("Cannot move there - Try Again");
			throw new CommandException("Cannot move into Wall"); // can't move into a wall or player
		}
		// Costs one action point
		player.decrementAp();

		// Move the player
		player.setLocation(location);

		return;
	}
	
	/**
	 * Returns the current message to the client. Note that this becomes
	 * important when using multiple clients across a network, where this could
	 * send to multiple players
	 * 
	 * @param message
	 *            The message to be shouted Thread The thread that has shouted
	 */
	public void clientShout(String message, Player player) {
		for (int user = 0; user < userList.size(); user++) {
			// if its not my thread
			if (userList.get(user) != player) {
				// send the message to their threads
				userList.get(user).sendMessage(message);
			}
		}
	}
	
	/**
	 * Handles the client message ATTACK
	 * 
	 * Note: In the single player version of the game this doesn't do anything
	 * 
	 * @param direction
	 *            The direction in which to attack
	 * @return A message indicating the success or failure of the attack
	 * @throws CommandException
	 */
	public void clientAttack(CompassDirection direction, Player player) throws CommandException {
		assertPlayerExists(player);
		ensureNoWinner();
		try {
			assertPlayerAP(player);
		} catch (IllegalStateException e) {
			throw new CommandException(e.getMessage());
		}

		/**
		 * TODO .... This code does not need to be filled in until Coursework 3!
		 * 
		 * 1. Work out which player the attack is on...
		 * 
		 * 2. Have you hit the target? - hint, you might want to make the chance
		 * of a successful attack 75%?
		 * 
		 * 2.1 if the player has hit the target then hp of the target should be
		 * reduced based on this formula...
		 * 
		 * damage = 1 + player.sword - target.armour
		 * 
		 * i.e. the damage inflicted is 1 + 1 if the player has a sword and - 1
		 * if the target has armour.
		 * 
		 * Player and target are informed about the attack as set out in the
		 * wire_spec
		 * 
		 * 2.2 if the player has missed the target then nothing happens.
		 * Optionally, the player and target can be informed about the failed
		 * attack
		 * 
		 */
		if (player.remainingAp() == 0) {
			// Force the end of turn
			clientEndTurn(player);
		}

		throw new CommandException("attacking (" + direction.toString() + ") has not been implemented");
	}
	
	public void clientEndTurn(Player player) {
		// check they exsist
		assertPlayerExists(player);
		// end turn to be sent to output
		player.endTurn();
		// see who goes next
		checkNewTurn(player);
	}
	
	/**
	 * Handles the client message PICKUP. Generally it decrements AP, and gives
	 * the player the item that they picked up Also removes the item from the
	 * map
	 * 
	 * @return A message indicating the success or failure of the action of
	 *         picking up.
	 * @throws CommandException
	 */
	public synchronized void clientPickup(Player player) throws CommandException {
		assertPlayerExists(player);
		ensureNoWinner();
		try {
			assertPlayerAP(player);
		} catch (IllegalStateException e) {
			throw new CommandException(e.getMessage());
		}

		final Tile playersTile = player.getTempTile();

		// Check that there is something to pick up
		if (!playersTile.hasItem()) {
			throw new CommandException("nothing to pick up");
		}

		// Get the item
		final GameItem item = playersTile.getItem();

		if (player.hasItem(item)) {
			throw new CommandException("already have item");
		}

		// remove one ap for pickup
		player.decrementAp();

		player.giveItem(item);
		playersTile.removeItem();

		// check to see if this has changed someones map
		checkChange(player);

		if (player.remainingAp() == 0) {
			// Force the end of turn
			clientEndTurn(player);
		}
	}
	
	/**
	 * If the Players field of View is Changed then tell them via "CHANGE"
	 * 
	 * @return A message indicating the success or failure of the action of
	 *         picking up.
	 * @throws CommandException
	 */
	public void checkChange(Player player) {
		for (int user = 0; user < userList.size(); user++) {
			// not my thread
			if ((userList.get(user) != player)) {
				// if my view is different from look
				if (userList.get(user).getLookView() != clientLook(userList.get(user))) {
					// tell the user change and get the new look for LOOKView
					System.out.println("change");
					userList.get(user).setLookView(clientLook(userList.get(user)));
					userList.get(user).change();
				}
			}
		}
	}
	
	/**
	 * Sets the player's position. This is used as a cheating or debug command.
	 * It is particularly useful for testing, as it gets rounds the randomness
	 * of the player start position.
	 * 
	 * @param col
	 *            the column of the location to put the player
	 * @param row
	 *            the row to location to put the player
	 * @throws CommandException
	 */
	public synchronized void setPlayerPosition(int col, int row, Player player) throws CommandException {
		assertPlayerExists(player);
		final Location location = new Location(col, row);

		if (!map.insideMap(location)) {
			throw new CommandException("invalid position");
		}

		if (!map.getMapCell(location).isWalkable()) {
			throw new CommandException("cannot walk on this tile");
		}

		player.setLocation(location);
	}
	
	/**
	 * Passes the goal back
	 * 
	 * @return the current goal
	 */
	public synchronized int getGoal() {
		return map.getGoal();
	}
	
	/**
	 * Once a player has performed an action the game needs to move onto the
	 * next turn to do this the game needs to check for a win and then test to
	 * see if the current player has more AP left.
	 * 
	 */
	public synchronized void advanceTurn(Player player) {
		// Check if the player has won
		if ((player.getGold() >= map.getGoal()) && (player.getTempTile().isExit())) {
			// Player should not be able to move if they have won
			assert (!this.playerWon);

			this.playerWon = true;
			player.win();

			for (int user = 0; user < userList.size(); user++) {
				// tell all other threads they have lost
				if (userList.get(user) != player) {
					userList.get(user).lose();
				}
			}

		} else {
			if (player.isDead()) {
				// SEND MESSAGE THEY HAVE DIED
				die(player);
			}

			if (player.remainingAp() == 0) {
				// Force the end of turn
				clientEndTurn(player);
			}
		}
	}

	/**
	 * Removes the player form the map by changing their 'P' to the value
	 * underneath them
	 */
	public void removeplayerfromMap(Player player) {
		map.removeUser(player.getLocation(), player);
	}
	
	public ArrayList<String> getUserNames() {
		return userNames;
	}

	public void setUserNames(String name) {
		userNames.add(name);
	}

	/**
	 * Ensures a player has enough AP, otherwise a runtime error is raised,
	 * since the turn should have been advanced. In a multiplayer example, this
	 * is still a bug, since the server should have checked whose turn it was.
	 * 
	 * @throws RuntimeException
	 */
	private void assertPlayerAP(Player player) throws RuntimeException {
		if (player.remainingAp() == 0) {
			throw new IllegalStateException("Player has 0 ap");
		}
	}
	
	

	/**
	 * Ensure that no player has won the game. Throws a CommandException if
	 * someone has one, preventing the command from executing
	 * 
	 * @throws CommandException
	 */
	private void ensureNoWinner() throws CommandException {
		if (this.playerWon) {
			throw new CommandException("the game is over");
		}
	}

	/**
	 * Ensures a player has been added to the map. Otherwise, an exception is
	 * raised. In a multiplayer scenario, this could ensure a player by given ID
	 * exists.
	 * 
	 * @throws RuntimeException
	 */
	private void assertPlayerExists(Player player) throws RuntimeException {
		if (player == null) {
			throw new IllegalStateException(": Player has not been added.");
		}
	}

	/**
	 * Removes the player form the map by changing their 'P' to the value
	 * underneath them
	 */
	private void removePlayerFromMap(Player player) {
		map.removeUser(player.getLocation(), player);
	}

	/**
	 * This doesn't really do anything as yet, other than reset the player AP.
	 * It should do more when the game is multiplayer
	 * 
	 * @param thread
	 */
	private void startTurn(Player player) {
		player.startTurn();
	}

	/**
	 * Generates a randomised start location
	 * 
	 * @return a random location where a player can start
	 */
	private Location generateRandomStartLocation() {
		if (!atLeastOneNonWallLocation()) {
			throw new IllegalStateException("There is no free tile available for the player to be placed");
		}
		while (true) {
			// Generate a random location
			final Random random = new Random();
			final int randomRow = random.nextInt(map.getMapHeight());
			final int randomCol = random.nextInt(map.getMapWidth());
			final Location location = new Location(randomCol, randomRow);
			if (map.getMapCell(location).isWalkable()) {
				// If it's not a wall then we can put them there
				return location;
			}
		}
	}

	/**
	 * Searches a possible tile to use by the player, i.e. non-wall. The map is
	 * traversed from (0,0) to (maxY,MaxX).
	 * 
	 * @return true if there is at least one non-wall location, false otherwise
	 */
	private boolean atLeastOneNonWallLocation() {
		for (int x = 0; x < map.getMapWidth(); x++) {
			for (int y = 0; y < map.getMapHeight(); y++) {
				if (map.getMapCell(new Location(x, y)).isWalkable()) {
					// If it's not a wall then we can put them there
					return true;
				}
			}
		}
		return false;
	}

}
