package dod.game;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;

import dod.ClientThread;
import dod.game.items.Gold;

/**
 * Class containing the map used by the game engine. Allows for reading in ASCII
 * art maps from a file, or just using the default map.
 */
public class Map {
	// The name of the map
	private String name;

	private List<String> lines;

	// The tiles of the map, stored in row-major order, i.e. [row][col]
	private Tile map[][];

	// The number of gold required to win
	private int goal;

	// The lines containing the name and goal, and rest of the map
	private static final int NAMELINE = 0;
	private static final int GOALLINE = 1;
	private static final int MAPBEGINLINE = 2;

	// Minimum number of lines
	private static final int MINLINES = 3;

	/**
	 * Creates a map from the file specified. Note that this is not robust...
	 * 
	 * @param filename
	 *            The name of the file to load the map from
	 * @throws ParseException
	 * @throws FileNotFoundException
	 */
	public Map(String filename) throws ParseException, FileNotFoundException {
		lines = readFile(filename);

		if (lines.size() < MINLINES) {
			throw new ParseException("a map file must contain at least three lines", lines.size());
		}

		// The first line should always be the name of the map.
		parseMapName(lines.get(NAMELINE));

		// The second line should be the goal.
		parseMapGoal(lines.get(GOALLINE));

		// Read in the map data from the file
		readMap(lines);
	}

	private void readMap(List<String> lines) throws ParseException, IllegalStateException {
		// Read the rest of the map
		final int mapWidth = lines.get(MAPBEGINLINE).length();
		final int mapHeight = lines.size() - MAPBEGINLINE;

		map = new Tile[mapHeight][mapWidth];

		for (int row = 0; row < mapHeight; row++) {
			final int lineNum = row + MAPBEGINLINE;
			final String line = lines.get(lineNum);

			if (line.length() != mapWidth) {
				throw new ParseException("all lines must be the same length", lineNum);
			}

			for (int col = 0; col < line.length(); col++) {
				// Just use the character representation in the input file.

				try {
					map[row][col] = Tile.fromChar(line.charAt(col), 1);
				} catch (final IllegalArgumentException e) {
					throw new ParseException("Invalid character (col:" + col + ")", lineNum);
				}
			}
		}
	}

	private void parseMapGoal(String secondLine) throws ParseException {
		final String goalString = getStringAfterTag(secondLine, "win", GOALLINE);
		try {
			goal = Integer.parseInt(goalString);
		} catch (final NumberFormatException e) {
			throw new ParseException("map goal should be an integer", GOALLINE);
		}
	}

	private void parseMapName(String firstLine) throws ParseException {
		name = getStringAfterTag(firstLine, "name", NAMELINE);
	}

	private List<String> readFile(String filename) throws FileNotFoundException {
		Scanner scanner = null;
		lines = new ArrayList<String>(); // final List<String>
		try {
			// Use a scanner to read all the lines
			scanner = new Scanner(new FileReader(filename));
			// Store all the lines in the file
			while (scanner.hasNextLine()) {
				lines.add(scanner.nextLine());
			}
		} finally {
			if (scanner != null) {
				scanner.close();
			}
		}
		return lines;
	}

	private String getStringAfterTag(String line, String tag, int lineNum) throws ParseException {
		// The first space will be after the tag
		final int firstSpace = line.indexOf(" ");

		// There may not be a first space
		if (firstSpace == -1) {
			throw new ParseException(
					tag + " not specified in file; the " + tag + " should be be preceded with \"" + tag + "\"",
					lineNum);
		}

		// Check that the tag is the first "word"
		if (line.substring(0, firstSpace).equals(tag)) {
			return line.substring(firstSpace + 1);
		} else {
			throw new ParseException("The map" + tag + "should be preceded with \"" + tag + "\"", lineNum);
		}
	}
	
	/**
	 * @return The width of the map
	 */
	public int getMapWidth() {
		return map[0].length;
	}

	/**
	 * @return The height of the map
	 */
	public int getMapHeight() {
		return map.length;
	}
	
	/**
	 * 
	 * @return The amount of gold required to win on this map
	 */
	public int getGoal() {
		return goal;
	}
	
	/**
	 * @return The name of the map
	 */
	public String getName() {
		return name;
	}

	/**
	 * Returns the contents of the cell of the map at the location
	 * 
	 * @return the contents of the cell
	 */
	public Tile getMapCell(Location location) {
		return map[location.getRow()][location.getCol()];
	}
	
	/**
	 * Will place the player in the map, Using 'P' and hold the value of what is
	 * beneath it
	 * 
	 * @param thread
	 *            The current user Location The location of the User
	 */
	public void playerPlace(Location location, Player player) {
		// value before 'P'
		player.setTempTile(map[location.getRow()][location.getCol()]);
		// location before 'P'
		player.setTempVal(location);
		// Value is now 'P'
		map[location.getRow()][location.getCol()] = Tile.fromChar('P', 1);
		// your current positon's tile
		player.setCurrentTile(map[location.getRow()][location.getCol()]);
	}

	/**
	 * Will update the 'P' on the map and the value beneath it
	 * 
	 * @param thread
	 *            The current user Location The location of the User
	 */
	public void updatePlayer(Location location, Player player) {
		// set the previous tile to what it was before now the player has moved 
		map[player.getTempVal().getRow()][player.getTempVal().getCol()] = player.getTempTile();
		// same as above 
		player.setTempTile(map[location.getRow()][location.getCol()]);
		player.setTempVal(location);
		map[location.getRow()][location.getCol()] = Tile.fromChar('P', player.getGold());
		player.setCurrentTile(map[location.getRow()][location.getCol()]);
	}

	/**
	 * Will remove the player from the map by setting it to the value underneath
	 * 
	 * @param thread
	 *            The current user Location The location of the User
	 */
	public void removeUser(Location location, Player player) {
		map[location.getRow()][location.getCol()] = player.getTempTile();
	}

	/**
	 * Used to check if a location is a valid location in the map
	 * 
	 * @return true if the location is valid, false otherwise
	 */
	public boolean insideMap(Location location) {
		if ((location.getCol() < 0) || (location.getCol() >= getMapWidth()) || (location.getRow() < 0)
				|| (location.getRow() >= getMapHeight())) {
			return false;
		}
		return true;
	}


	/**
	 * @return the amount of gold that has not been picked up on the map
	 */
	public int remainingGold() {
		int goldCount = 0;

		for (final Tile[] tileRow : map) {
			for (final Tile tile : tileRow) {
				if (tile.hasItem()) {
					if (tile.getItem().getClass() == Gold.class) {
						goldCount++;
						// Should actually get the gold and then return the value and add to it
					}
				}
			}
		}
		return goldCount;
	}
	
	/**
	 * Places Gold of a certain value on the current location tile
	 */
	public void setMapCellGold(Location location, int gold) {
		map[location.getRow()][location.getCol()] = Tile.fromChar('G', gold);
	}


}
