package dod.game;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;

import dod.ClientThread;
import dod.game.items.GameItem;
import dod.game.items.GameItemConsumer;

/**
 * Class representing a player
 * 
 * Note: Player variables should be kept as a part of the server functionality
 * they are the game's internal representation of a player not used by the
 * client side code.
 */
public class Player implements GameItemConsumer {
	private String name;

	// Does the player have a default name
	private boolean defaultName = true;

	// The player may be "listened to" to interpret updates
	private final PlayerListener listener;
	private ClientThread thread;
	private Map map;

	// Location on the map
	private Location location;

	private Tile tempTile;
	private Location tempVal;
	private Tile currentTile;

	// How much gold they have, initially zero
	private int gold = 0;

	// Player attribute value things
	private int hp = 3;
	private int ap = 0;

	// Items the player has
	private ArrayList<GameItem> items;

	// current look reply
	private String lookView;

	// Constants
	// How many AP does a player have by default
	private static final int defaultAP = 6;

	// How many AP does the player lose per item
	private static final int apPenaltyPerItem = 1;

	// How far can a player see by default and with a lantern
	private static final int defaultLookDistance = 2;

	/**
	 * Constructor for players
	 * 
	 * @param name
	 *            the name of the player
	 * @param location
	 *            the location of the player
	 * @param listener
	 *            a player may be "listened to" for updates.
	 */
	public Player(String name, Location location, PlayerListener listener, ClientThread thread, Map map) {
		this.name = name;
		this.location = location;
		this.map = map;
		this.thread = thread;
		// By default the player starts with nothing
		this.items = new ArrayList<GameItem>();
		this.listener = listener;
	}

	/**
	 * SET LOCATION IN MAP TO "P"
	 * 
	 * @param thread
	 *            - current user
	 * @param map
	 *            - current map
	 */
	public void placeThePlayer() {
		map.playerPlace(location, this);
	}

	/**
	 * Update Player Location in map
	 * 
	 * @param thread
	 *            - current user
	 * @param map
	 *            - current map
	 */
	public void updateThePlayer() {
		map.updatePlayer(location, this);
	}

	/**
	 * Returns the current location of the player
	 * 
	 * @return player's location
	 */
	public Location getLocation() {
		return location;
	}

	/**
	 * Sets the player to a new location
	 * 
	 * @param location
	 *            the player's new location
	 */
	public void setLocation(Location location) {
		this.location = location;
		updateThePlayer();
	}

	/**
	 * Returns the name of the player
	 * 
	 * @return The name of the player
	 */
	public String getName() {
		return name;
	}

	/**
	 * Sets the name of the player
	 * 
	 * @param name
	 *            The new name of the player
	 * @throws CommandException
	 */
	public void setName(String name, GameLogic game) throws CommandException {
		if (!defaultName) {
			throw new CommandException("player's name already set");
		}
		if (game.getUserNames().contains(name)) {
			throw new CommandException("player's name in use");
		} else {
			game.getUserNames().remove(this.name);
			this.name = name;
			game.setUserNames(name);
			defaultName = false;
		}
	}

	/**
	 * @return The amount of gold the player has
	 */
	public int getGold() {
		return gold;
	}

	/**
	 * Adds gold to the player
	 * 
	 * @param gold
	 *            Amount of gold to give to the player
	 */
	@Override
	public void addGold(int gold) {
		this.gold += gold;
		this.listener.treasureChange(gold);
	}

	/**
	 * A player is dead if they have 0hp or less
	 * 
	 * @return true if the player is dead, false otherwise
	 */
	public boolean isDead() {
		return (hp <= 0);
	}

	/**
	 * @return The amount of HP the player has
	 */
	public int getHp() {
		return hp;
	}

	/**
	 * Increments the player's health, e.g. for health potion
	 * 
	 * @param hp
	 *            The amount of hp to add ot the player
	 */
	@Override
	public void incrementHealth(int hp) {
		this.hp += hp;
		this.zeroAP();
		this.listener.hpChange(hp);
	}

	/**
	 * Returns the amount of AP the player has
	 * 
	 * @return The amount of AP the player has
	 */
	public int remainingAp() {
		return ap;
	}

	/**
	 * Decreases the amount of AP the player has by 1
	 */
	public void decrementAp() {
		ap--;
		assert ap >= 0;
	}

	/**
	 * Reset the player's AP to the initial value.
	 */
	public void resetAP() {
		ap = initialAP();
	}

	/**
	 * Calculates the number of AP a player starts his or her turn with
	 * 
	 * @return The amount of AP at the start of a turn
	 */
	private int initialAP() {
		final int initialAP = defaultAP - apPenaltyPerItem * items.size();
		if (initialAP < 0) {
			// Better drop some items
			return 0;
		}
		return initialAP;
	}

	/**
	 * Set's the player's AP to zero instantly
	 */
	@Override
	public void zeroAP() {
		ap = 0;
	}

	/**
	 * Calculates the distances the player can see
	 * 
	 * @return the distance visible to the player
	 */
	public int lookDistance() {
		int lookDistance = defaultLookDistance;
		// Some items, e.g. the lantern, may increase the look distance
		for (final GameItem item : items) {
			lookDistance += item.lookDistanceIncrease();
		}
		return lookDistance;
	}

	/**
	 * Check if the player already has a given item type (e.g. any sword, not
	 * just "that" sword)
	 * 
	 * @param item
	 *            An instance of the item to compare with
	 * @return true if the player has the item, false otherwise
	 */
	public boolean hasItem(GameItem item) {
		for (final GameItem itemToCompare : items) {
			if (item.getClass() == itemToCompare.getClass()) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Gives the item to the player. The player allows the item to act on the
	 * player (through this), and keeps it in the item list if it is retainable.
	 * 
	 * @param item
	 *            the item to pick up
	 */
	public void giveItem(GameItem item) { 
		if (hasItem(item)) {
			throw new IllegalStateException("the player already has this item.");
		}
		// The item may do something to the player straight away
		item.processPickUp(this);
		// See if the item is retained by the player
		if (item.isRetainable()) {
			items.add(item);
		}
	}

	/**
	 * @param message
	 *            message to send to the listener
	 */
	public void sendMessage(String message) {
		listener.sendMessage(message);
	}

	/**
	 * @param Message
	 *            - HTML to be sent
	 * @param ListUser
	 *            - User to be sent message
	 * @param Currentthread
	 *            - User message was sent from
	 */
	public void sendChatMessage(String Message) {
		// get Time
		Date dNow = new Date();
		SimpleDateFormat ft = new SimpleDateFormat("hh:mm:ss");
		// turn time into a string
		String timeStamp = ft.format(dNow);
		// add the Time Name of player and message into string to be sent as
		// HTML
		String NameAndTime = "<font>" + timeStamp + " " + name + ": " + "</font>";
		thread.sendChatMessage(NameAndTime + Message);
	}

	/**
	 * Handle the start of a player's turn
	 */
	public void startTurn() {
		resetAP();
		listener.startTurn();
	}

	/**
	 * Handle the end of a player's turn
	 */
	public void endTurn() {
		zeroAP();
		listener.endTurn();
	}

	/**
	 * Handle the player winning
	 */
	public void win() {
		listener.win();
	}

	public void lose() {
		listener.lose();
	}

	@Override
	public String toString() {
		return "Player";
	}

	@Override
	public char toChar() {
		return 'P';
	}

	/**
	 * Tell the listener that a change has happened
	 * 
	 * @param thread
	 *            - current User
	 */
	public void change() {
		listener.change();
	}

	/**
	 * Returns true if a player can see a tile, based on the offset from the
	 * player
	 * 
	 * @param rowOffset
	 * @param colOffset
	 * @return true if the player can see the tile with the specified offset.
	 */
	public boolean canSeeTile(int rowOffset, int colOffset) {
		// This is based on the Manhattan distance
		final boolean canSeeTile = (Math.abs(rowOffset) + Math.abs(colOffset) <= lookDistance() + 1);
		return canSeeTile;
	}

	public Tile getTempTile() {
		return tempTile;
	}

	public void setTempTile(Tile tempTile) {
		this.tempTile = tempTile;
	}

	public Location getTempVal() {
		return tempVal;
	}

	public void setTempVal(Location tempVal) {
		this.tempVal = tempVal;
	}

	public Tile getCurrentTile() {
		return currentTile;
	}

	public void setCurrentTile(Tile currentTile) {
		this.currentTile = currentTile;
	}

	public String getLookView() {
		return lookView;
	}

	public void setLookView(String lookView) {
		this.lookView = lookView;
	}

}